package at.reisinger.server.objects;

import at.reisinger.*;
import at.reisinger.Package;
import at.reisinger.server.helper.JsonHelper;
import at.reisinger.server.helper.PostgresHelper;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.postgresql.shaded.com.ongres.scram.common.bouncycastle.pbkdf2.Pack;

import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Base64;
import java.util.List;
import java.util.Map;

/**
 * Erstellt und sendet eine Response anhand des Requests
 * @author Georg Reisinger
 */
public class Response {

    //private final Socket socket;
    private final int          id;
    private PrintStream out;
    private String       cmd;
    private String       url;
    private String payload;
    private String authString;

    /**
     * Nimmt die Daten des requests und generiert eine Response
     * @param id Thread Id
     * @param url Request Url
     * @param cmd Request CMD
     * @param out out Print Stream
     * @param authString Der MsgHandler
     * @param payload Payload des Requests
     */
    public Response(int id, String url, String cmd, PrintStream out, String authString, String payload){
        this.id = id;
        this.authString = authString;
        this.url = url;
        this.cmd = cmd;
        this.out = out;
        this.payload = payload;
        if (this.url != null) {
            if (this.cmd.equals("GET")) {
                if(login()) {
                    getMethodes();
                }else{
                    sendError("401");
                }
            }else if (this.cmd.equals("POST")){
                try {
                    postMethodes();
                } catch (JsonProcessingException e) {
                    e.printStackTrace();
                }
            }else if (this.cmd.equals("PUT")){
                if(login()) {
                    putMethodes();
                }else{
                    sendError("401");
                }
            }else if (this.cmd.equals("DELETE")){
                if(login()) {
                    deleteMethodes();
                }else{
                    sendError("401");
                }
            }else{
                sendError("405");
            }
        }
    }

    /**
     * Alle Get Methoden
     * Nur für angemeldete user erreichbar
     */
    private void getMethodes(){
        if (this.url.startsWith("/users")) {
            String username = this.url.substring(this.url.lastIndexOf('/') + 1);
            System.out.println("Username in /users/{username}: " + username);
            User user = null;
            try {
                user = new DBConnection().getUser(username);
            } catch (SQLException throwables) {
                throwables.printStackTrace();
                sendError("401");
            }
            if (user != null){
                String userJson = JsonHelper.userToJson(user);
                if(userJson != null && !userJson.isEmpty()){
                    sendResponse(userJson, "200");
                }
            }else{
                sendError("500");
            }
        } else if (this.url.startsWith("/cards")) {
            String username = basicAuthGetUsername(this.authString);
            Cards allCards = new DBConnection().getCards(username);
            String jsonCards = JsonHelper.objToJson(allCards);
            if (jsonCards != null && !jsonCards.isEmpty()){
                sendResponse(jsonCards, "200");
            }else{
                sendError("500");
            }
        }else if(this.url.startsWith("/deck")) {
            String format = this.url.substring(this.url.lastIndexOf('?') + 1);
            System.out.println("Format in /deck?{format=plain}: " + format);
            String username = basicAuthGetUsername(this.authString);
            ArrayList<String> allCards = new DBConnection().getDeck(username);
            Cards deck = null;
            Object tmp = null;
            if(format.startsWith("format=plain")){
                tmp = allCards;
            }else{
                deck = new DBConnection().getCardsFromIDList(allCards);
                tmp = deck;
            }
            String jsonCards = JsonHelper.objToJson(tmp);
            if (jsonCards != null && !jsonCards.isEmpty()) {
                sendResponse(jsonCards, "200");
            } else {
                sendError("500");
            }
        }
    }

    /**
     * Vergleicht das angegebene PW mit dem gespeicherten PW, True wenn richtig
     * @param username User aus DB
     * @param password PW zu User aus DB
     * @param compareTo Base64 String von Basic Auth
     * @return True if Login is correct, false if incorrect
     */
    private boolean basicAuth(String username, String password, String compareTo){
        String authStringEnc = basicAuthBase64(username, password);
        if (compareTo.equals(authStringEnc)){
            return true;
        }
        return false;
    }

    /**
     * Erstellt den basicAuth Base64 Token aus Username und Passwort
     * @param username Username für den Token
     * @param password Passwort für den Token
     * @return Der Base64 BasicAtuth String
     */
    private String basicAuthBase64(String username, String password){
        String authString = username + ":" + password;
        byte[] authEncBytes = Base64.getEncoder().encode(authString.getBytes());
        String authStringEnc = new String(authEncBytes);
        return authStringEnc;
    }

    /**
     * Aus dem Base64 BasicAuth String wird der Username extrahiert
     * @param authString Base64 BasicAuth String aus dem der Username ausgelesen werden soll
     * @return Der Username der in dem Base64 BasicAuth String ist
     */
    private String basicAuthGetUsername(String authString) {
        String[] values;
        if (authString != null) {
            byte[] credDecoded = Base64.getDecoder().decode(authString);
            String credentials = new String(credDecoded);
            values = credentials.split(":", 2);
        }else{
            values = null;
        }
        return values[0];
    }

    /**
     * Alle Post Methoden, siehe API beschreibung für genauere details
     *
     * Kann auch von nicht angemeldeten usern erreicht werden.
     * login() wird hier verwendet
     * @throws JsonProcessingException Fehler beim de-/serialisieren
     */
    private void postMethodes() throws JsonProcessingException {
        if (this.url.startsWith("/users")) {
            Map<String, Object> map = JsonHelper.jsonPayloadToMap(this.payload);
            String username = (String) map.get("Username");
            String password = (String) map.get("Password");
            User newUser = new User(new Credentials(username, password), username, username, "not implemented",  new Coins(20), "BIO", "IMAGE");
            System.out.println("CON DB");
            DBConnection con = new DBConnection();
            if(!con.addUser(newUser.getCredentials().getUsername(), newUser.getCredentials().getPasswort(), newUser.getCredentials().getUsername(), newUser.getEmail(), newUser.getBio(), newUser.getImage())){
                sendError("409");
            }
            String userJson = JsonHelper.userToJson(newUser);;
            if(userJson != null) {
                sendResponse(userJson, "201");
            }else{
                sendError("500");
            }
        }else if (this.url.startsWith("/sessions")) {
            Map<String, Object> map = JsonHelper.jsonPayloadToMap(this.payload);
            String username = (String) map.get("Username");
            String password = (String) map.get("Password");
            User user = null;

            DBConnection con = new DBConnection();
            try {
                user = con.getUser(username);
            } catch (SQLException throwables) {
                throwables.printStackTrace();
            }

            String authString = basicAuthBase64(user.getCredentials().getUsername(), user.getCredentials().getPasswort());
            if(basicAuth(username, password, authString)){
                sendResponse(authString, "200");
            }else{
                sendError("401");
            }
        }else if (this.url.startsWith("/packages")) {
            if (basicAuth("admin", "istrator", this.authString)) {
                ObjectMapper objectMapper = new ObjectMapper();
                ArrayList<Card> listCards = objectMapper.readValue(this.payload, new TypeReference<ArrayList<Card>>() {
                });
                Package packageCards = new Package(new Cards(listCards), new DBConnection().nextPackageId() + "", 5);
                if (!new DBConnection().addPackage(packageCards)) {
                    sendError("500");
                } else {
                    String packageJson = JsonHelper.objToJson(packageCards);
                    if (packageJson != null) {
                        sendResponse(packageJson, "201");
                    } else {
                        sendError("500");
                    }
                }
            } else {
                sendError("401");
            }
        }else if (this.url.startsWith("/transactions/packages")) {
            if (login()) {
                DBConnection db = new DBConnection();
                Package newPackage = db.userAcquirePackage(basicAuthGetUsername(this.authString));
                if (newPackage == null) {
                    sendError("500");
                } else {
                    String packageJson = JsonHelper.objToJson(newPackage);
                    if (packageJson == null) {
                        sendError("500");
                    } else {
                        sendResponse(packageJson, "200");
                    }
                }
            }else{
                sendError("401");
            }
        }else if (this.url.startsWith("/tradings")) {
            if(login()){
                Map<String, Object> map = JsonHelper.jsonPayloadToMap(this.payload);
                String id = (String)map.get("Id");
                String cardtotrade = (String)map.get("CardToTrade");
                String type = (String)map.get("Type");
                double mindamage = (double)map.get("MinimumDamage");
                ArrayList<String> cardname = new ArrayList<String>();
                cardname.add(cardtotrade);
                Cards cards = new DBConnection().getCardsFromIDList(cardname);
                Card cardtoTradeC = null;
                if (cards != null){
                    cardtoTradeC = cards.getCards().get(0);
                    if (cardtoTradeC != null){
                        TradingDeal tradingDeal = new TradingDeal(id, cardtoTradeC, mindamage, type, basicAuthGetUsername(this.authString));
                        if (tradingDeal != null){
                            String tradingJson = JsonHelper.objToJson(tradingDeal);
                            if (tradingDeal != null){
                                //DB Trading deal add
                                // Lock Card for deck
                                    //Wenn card im deck, dann del aus deck
                                //new DBConnection().;
                                sendResponse(tradingJson, "201");
                            }else{
                                sendError("500");
                            }
                        }else{
                            sendError("500");
                        }
                    }else {
                        sendError("500");
                    }
                }else{
                    sendError("500");
                }

            }else{
                sendError("401");
            }
        }
    }

    /**
     * Prüft den Login, anhand des BasicAuth Base64 String
     * @return True wenn login daten korrekt, sonst false
     */
    private boolean login(){
        String username = basicAuthGetUsername(this.authString);
        User user = null;
        DBConnection con = new DBConnection();
        try {
            user = con.getUser(username);
        } catch (SQLException throwables) {
            throwables.printStackTrace();
            return false;
        }
        if (basicAuth(user.getCredentials().getUsername(), user.getCredentials().getPasswort(), this.authString)){
            return true;
        }else{
            return false;
        }
    }

    /**
     * Alle Put Methoden
     * Nur für angemeldete user erreichbar
     * @throws JsonProcessingException Fehler beim de-/serialisieren
     */
    private void putMethodes(){
        if (this.url.startsWith("/users")) {
            String username = this.url.substring(this.url.lastIndexOf('/') + 1);
            System.out.println("Username in /users/{username}: " + username);
            User user = null;
            try {
                user = new DBConnection().getUser(username);
            } catch (SQLException throwables) {
                throwables.printStackTrace();
                sendError("401");
            }
            if(user != null) {
                Map<String, Object> map = JsonHelper.jsonPayloadToMap(this.payload);
                String bio = (String) map.get("Bio");
                String image = (String) map.get("Image");
                String name = (String) map.get("Name");
                System.out.println("Request /users:");
                System.out.println("Bio: " + bio + "\nImage: " + image + "\nName: " + name);
                user.setBio(bio);
                user.setImage(image);
                user.setNachname(name);
                if (new DBConnection().updateUser(username, user.getBio(), user.getImage(), user.getNachname())) {
                    sendResponse(JsonHelper.userToJson(user), "200");
                } else {
                    sendError("500");
                }
            }else{
                sendError("500");
            }
        }else if(this.url.startsWith("/deck")) {
            List<String> deckIds = JsonHelper.jsonPayloadToList(this.payload);
            if (deckIds != null){
                if (new DBConnection().setDeck(basicAuthGetUsername(this.authString), deckIds)){
                    Cards deck = new DBConnection().getCardsFromIDList(deckIds);
                    String deckJson = JsonHelper.objToJson(deck);
                    if (deck != null && deckJson != null){
                        sendResponse(deckJson, "200");
                    }else {
                        sendError("500");
                    }
                }else {
                    sendError("500");
                }
            }else{
                sendError("500");
            }
        }
    }



    /**
     * Alle Delete Methoden, siehe API beschreibung für genauere details
     * Nur für angemeldete user erreichbar
     * @throws JsonProcessingException Fehler beim de-/serialisieren
     */
    private void deleteMethodes(){
        if (this.url.startsWith("/messages")) {

        }
    }


    /**
     * Sendet einen Error Response
     * @param errorCode Der Error Code
     */
    private void sendError(String errorCode) {
        out.print("HTTP/1.0 "+errorCode+"\r\n");
        out.print("Server: Apache/0.8.4\r\n");
        out.print("Content-Type: text/plain\r\n");
        out.print("Content-Length: 1\r\n");
        out.print("\r\n");
        //out.print(responseText);
    }

    /**
     * Sendet eine Startseite, mit allen befehlen und wie man diese verwendet

    private void startseite() {
        sendResponse("lists all messages:   GET    /messages<br>" +
                "add message:          POST   /messages     (Payload: the message; Response an id like1)<br>" +
                "show first message:   GET    /messages/1<br>" +
                "show third message:   GET    /messages/3<br>" +
                "update first message: PUT    /messages/1   (Payload: the message)<br>" +
                "remove first message: DELETE /messages/1<br>", "200");
    }
     */

    /**
     * Sendet eine Response
     * @param responseText Text der zu senden ist
     * @param code Http code
     */
    private void sendResponse(String responseText, String code){
        out.print("HTTP/1.0 "+code+"\r\n");
        out.print("Server: Apache/0.8.4\r\n");
        out.print("Content-Type: text/plain\r\n");
        out.print("Content-Length: "+responseText.length()+"\r\n");
        out.print("\r\n");
        out.print(responseText);
    }
}
